#
# Copyright 2014 Xavier Bruhiere
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import logbook

import intuition.data.remote as remote


class PortfolioFactory():
    '''
    Manages portfolio during simulation, and stays aware of the situation
    through the update() method.

    User strategies call it with a dictionnnary of detected opportunities (i.e.
    buy or sell signals).  Then the optimize function computes assets
    allocation, returning a dictionnary of symbols with their weigths or amount
    to reallocate.
                                  __________________________      _____________
    signals {'google': 745.5} --> |                         | -> |            |
                                  | trade_signals_handler() |    | optimize() |
    orders  {'google': 34}    <-- |_________________________| <- |____________|

    This is abstract class, inheretid class will eventally overwrite optmize()
    to expose their own asset allocation strategy.
    '''

    __metaclass__ = abc.ABCMeta

    # Zipline portfolio object, updated during simulation with self.date
    portfolio = None
    perfs = None
    date = None

    #TODO Add in the constructor or setup parameters some general settings like
    #     maximum weights, positions, frequency, ...
    #TODO Better to return 0 stocks to trade: remove the field
    #NOTE Regarding portfolio constraints: from a set of user-defined
    # parameters, a unique set should be constructed. Then the solution
    # provided by optimize function would have to be a subset of it. (classic
    # mathematical solution) Finally it should be defined how to handle
    # non-correct solutions
    def __init__(self, configuration):
        '''
        Parameters
            configuration : dict
                Named parameters used either for general portfolio settings
                (server and constraints), and for user optimizer function
        '''
        # Portfolio owner, mainly used for database saving and client
        # communication
        self.log = logbook.Logger('intuition.portfolio')

        # Other parameters are used in user's optimize() method
        self._optimizer_parameters = configuration

        # In case user optimization would need to retrieve more data
        self.data = remote.Data()
        self.initialize(configuration)

    def initialize(self, configuration):
        ''' Users should overwrite this method '''
        pass

    @abc.abstractmethod
    def optimize(self, date, to_buy, to_sell, parameters):
        ''' Users should overwrite this method '''
        pass

    def update(self, portfolio, date, perfs=None):
        '''
        Actualizes the portfolio universe
        ________________________________
        Parameters
            portfolio: zipline.Portfolio
                ndict object storing portfolio values at the given date
            date: datetime.datetime
                Current date in zipline simulation
        '''
        # Make the manager aware of current simulation
        self.portfolio = portfolio
        self.perfs = perfs
        self.date = date

    def trade_signals_handler(self, signals):
        '''
        Process buy and sell signals from the simulation
        ___________________________________________________________
        Parameters
            signals: dict
                hold stocks of interest, format: {"google": 0.8, "apple": -0.2}
                If the value is negative -> sell signal, otherwize buy one
                Values are ranged between [-1 1] regarding signal confidence
        ___________________________________________________________
        Return
            dict order_book, like {"google": 34, "apple": -56}
        '''
        order_book = {}

        # If value < 0, it's a sell signal on the key, else buy signal
        to_buy = dict(filter(lambda (sid, strength):
                             strength > 0, signals.iteritems()))
        to_sell = dict(filter(lambda (sid, strength):
                              strength < 0, signals.iteritems()))

        if not to_buy and not to_sell:
            # Nothing to do
            return {}

        # Compute the optimal portfolio allocation, using user defined function
        alloc, e_ret, e_risk = self.optimize(
            self.date, to_buy, to_sell, self._optimizer_parameters)

        #TODO Check about selling within available money
        #     and handle 250 stocks limit
        #TODO Handle max_* as well, ! already actif stocks

        ## Building orders for zipline
        #NOTE The follonwing in a separate function that could be used when
        #     catching message from user
        for t in alloc:
            # Handle allocation returned as number of stocks to order
            if isinstance(alloc[t], int):
                order_book[t] = alloc[t]

            # Handle allocation returned as stock weights to order
            elif isinstance(alloc[t], float):
                # Sell orders
                if alloc[t] <= 0:
                    order_book[t] = int(alloc[t] *
                                        self.portfolio.positions[t].amount)
                # Buy orders
                else:
                    # If we already trade this ticker, substract owned amount
                    # before computing number of stock to buy
                    if self.portfolio.positions[t].amount:
                        price = self.portfolio.positions[t].last_sale_price
                    else:
                        #FIXME Not clear if signals[t] is actually the price
                        price = signals[t]
                    order_book[t] = (int(alloc[t] *
                                     self.portfolio.portfolio_value / price)
                                     - self.portfolio.positions[t].amount)

        return order_book

    def advise(self, **kwargs):
        '''
        General parameters or user settings
        (maw_weigth, max_assets, max_frequency, commission cost)
        '''
        for name, value in kwargs.iteritems():
            self._optimizer_parameters[name] = value



#
# Copyright 2013 Xavier Bruhiere
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytz
import pandas as pd
import logbook

from zipline.finance.trading import TradingEnvironment
from zipline.utils.factory import create_simulation_parameters

import intuition.constants as constants
from intuition.data.utils import Exchanges
from intuition.data.loader import LiveBenchmark
from intuition.core.analyzes import Analyze
import intuition.utils.utils as utils


log = logbook.Logger('intuition.core.engine')


def _intuition_module(location):
    ''' Build the module path and import it '''
    path = location.split('.')
    obj = path.pop(-1)
    return utils.dynamic_import('.'.join(path), obj)


#NOTE Is there still a point to use here a constructor object instead of a
#     simple function ?
class TradingEngine(object):
    ''' Factory class wrapping zipline Backtester, returns the requested algo
    ready for use '''

    def __new__(self, identity, modules, strategy_conf):

        algo_obj = _intuition_module(modules['algorithm'])
        algo_obj.identity = identity
        trading_algo = algo_obj(properties=strategy_conf['algorithm'])

        trading_algo.set_logger(logbook.Logger('algo.' + identity))

        if modules['data']:
            trading_algo.set_data_generator(_intuition_module(modules['data']))

        # Use a portfolio manager
        if modules.get('manager'):
            log.info('initializing manager {}'.format(modules['manager']))
            # Linking to the algorithm the configured portfolio manager
            trading_algo.manager = _intuition_module(modules['manager'])(
                strategy_conf['manager'])
        else:
            trading_algo.manager = None
            log.info('no portfolio manager used')

        return trading_algo


class Simulation(object):
    ''' Setup and trigger trading sessions '''
    context = None

    def _get_benchmark_handler(self, last_trade):
        '''
        Setup a custom benchmark handler or let zipline manage it
        '''
        is_live = (last_trade > pd.datetime.now(pytz.utc))
        #NOTE minute hardcoded until more timedeltas supported
        return LiveBenchmark(
            last_trade, frequency='minute').surcharge_market_data \
            if is_live else None

    def configure_environment(self, last_trade, exchange):
        ''' Prepare benchmark loader and trading context '''

        # Setup the trading calendar from market informations
        if exchange in Exchanges:
            self.context = TradingEnvironment(
                bm_symbol=Exchanges[exchange]['symbol'],
                exchange_tz=Exchanges[exchange]['timezone'],
                load=self._get_benchmark_handler(last_trade))
        else:
            raise NotImplementedError('Because of computation limitation, \
                trading worldwide not permitted currently')

    def build(self, identity, modules, strategy=constants.DEFAULT_CONFIG):
        '''
        Wrapper of zipline run() method. Use the configuration set so far
        to build up the trading environment and launch the system
        '''
        self.engine = TradingEngine(identity, modules, strategy)

        self.initial_cash = strategy['manager'].get('cash', None)

    def run(self, data, auto=False):
        ''' wrap zipline.run() with finer control '''
        self.engine.auto = auto
        #FIXME crash if trading one day that is not a trading day
        with self.context:
            sim_params = create_simulation_parameters(
                capital_base=self.initial_cash,
                start=data['index'][0],
                end=data['index'][-1])

            daily_stats = self.engine.trade(data, sim_params=sim_params)

        return Analyze(
            results=daily_stats,
            metrics=self.engine.risk_report)



#!/usr/bin/env python
#
# Copyright 2014 Xavier Bruhiere
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from glob import glob

from setuptools import setup, find_packages

from intuition import __version__, __author__, __licence__


def get_requirements():
    with open('./requirements.txt') as requirements:
        # Avoid github based requirements and replace them
        deps = requirements.read().split('\n')[:-3]
        deps.append('zipline>=0.5.11.dev')
        deps.append('pandas>=0.13.0.dev')
        return deps


requires = [
    'beautifulsoup4>=4.3.2',
    'blist>=1.3.4',
    'Cython>=0.19.1',
    'Logbook>=0.6.0',
    'numpy>=1.8.0',
    'python-dateutil>=2.2',
    'pytz>=2013.8',
    'Quandl>=1.7',
    'requests>=2.0.1',
    'six>=1.4.1',
    'zipline>=0.5.11.dev',
    'pandas>=0.13.0.dev']


def long_description():
    try:
        #with codecs.open(readme, encoding='utf8') as f:
        with open('README.md') as f:
            return f.read()
    except IOError:
        return "failed to read README.md"


setup(
    name='intuition',
    version=__version__,
    description='A trading system building blocks',
    author=__author__,
    author_email='xavier.bruhiere@gmail.com',
    packages=find_packages(),
    long_description=long_description(),
    license=__licence__,
    install_requires=requires,
    url="https://github.com/hackliff/intuition",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: System :: Distributed Computing',
    ],
    scripts=['app/intuition'],
    data_files=[(os.path.expanduser('~/.intuition/data'), glob('./data/*'))],
    dependency_links=[
        'http://github.com/pydata/pandas/tarball/master#egg=pandas-0.13.0.dev',
        'http://github.com/quantopian/zipline/tarball/master#egg=zipline-0.5.11.dev']
)
